import cPickle as pickle
from cStringIO import StringIO

from pyramid.response import Response
from zope.interface import implementer, providedBy

from pyramid_caching.exc import SerializationError, DeserializationError
from pyramid_caching.interfaces import ISerializer, ISerializationAdapter

SERIALIZER_META_VERSION = 1
PICKLE_PROTOCOL = 2


def includeme(config):
    registry = config.registry

    utility = SerializerUtility(registry)
    registry.registerUtility(utility)

    utility.register_serialization_adapter(Response, ResponseAdapter)

    config.add_directive('get_serializer', get_serializer, action_wrap=False)


def get_serializer(config):
    return config.registry.getUtility(ISerializer)


@implementer(ISerializer)
class SerializerUtility(object):
    def __init__(self, registry):
        self.registry = registry

    def register_serialization_adapter(self, object_class, adapter_factory):
        # Register serializer by object class.
        self.registry.registerAdapter(lambda x: adapter_factory(),
                                      required=[object_class],
                                      provided=ISerializationAdapter,
                                      )
        # Register deserializer by type name.
        self.registry.registerAdapter(lambda x: adapter_factory(),
                                      required=[None],
                                      provided=ISerializationAdapter,
                                      name=adapter_factory.name,
                                      )

    def dumps(self, obj, adapter=None):
        if adapter is None:
            adapter = self.registry.queryAdapter(obj, ISerializationAdapter)
            if adapter is None:
                raise SerializationError(
                    "No encoder registered for %s" % providedBy(obj).__name__)

        meta = {
            'type': adapter.name,
            'version': SERIALIZER_META_VERSION,
            'payload': adapter.serialize(obj),
            }

        return pickle.dumps(meta, protocol=PICKLE_PROTOCOL)

    def loads(self, data):
        data = pickle.loads(data)
        if 'version' not in data or data['version'] != SERIALIZER_META_VERSION:
            return None
        adapter = self.registry.queryAdapter(None,
                                             ISerializationAdapter,
                                             name=data['type'])
        if adapter is None:
            raise DeserializationError("No decoder registered for type=%s",
                                       data['type'])
        return adapter.deserialize(data['payload'])


@implementer(ISerializationAdapter)
class ResponseAdapter(object):
    """Deserializer for Pyramid Response objects."""

    name = 'pyramid.response.Response'

    def serialize(self, response):
        return str(response)

    def deserialize(self, raw_response):
        res = Response.from_file(StringIO(raw_response))

        # Workaround for issue #99 in webob. All header names must be str
        # instances.
        res._headerlist = [(str(k), str(v)) for k, v in res._headerlist]

        return res



import os

from pyramid_caching.tests.functional import Base


class Test(Base):

    def setUp(self):
        os.environ['CACHING_ENABLED'] = 'false'

        # creating app to setup cache and decorate stuff.
        self.app
        from example.model import User, Session
        session = Session()
        session.add(User(id=1, name='Bob'))
        session.commit()

    def tearDown(self):
        os.environ.pop('CACHING_ENABLED', None)

    def test_disabled_api(self):
        # Check that part of the API is available (inactive)

        self.assertFalse(self.config.registry.settings['caching.enabled'])

        self.assertTrue(
            hasattr(self.config, 'register_sqla_session_caching_hook'))
        self.assertTrue(
            hasattr(self.config, 'register_sqla_base_class'))

    def test_disabled_cache_view(self):
        self.app.get('/users/1')
        self.app.get('/users/1')
        self.app.get('/users/1')



from zope.interface import Interface, Attribute


class IKeyVersioner(Interface):
    """
    Notes about the default values:

    There is two default values :
    - the return value of get()/get_multi() when the version doesn't exist
    - the stored version after an incr() operation

    Those default values are implementation specific and MUST be equivalent.

    See Redis implementation.
    """

    def get_multi(keys):
        """Get versions for a list of keys.

        Return a list of corresponding versions. Defaults is implementation
        specific.
        """

    def incr(key):
        """Increment an existing key version. Defaults is implementation
        specific.
        """


class IVersioner(Interface):

    def get_multi_keys(objects_or_classes):
        """Returns all versioned keys of each class or object in
        classes_or_objects.

        .. code-block:: ipython

            >>> from pyramid_caching.versioner import Versioner
            >>> from myapp.model import MyModel
            >>> versioner = Versioner()
            >>> res = versioner.get_class_keys([MyModel])
            >>> res == ['myapp.model.MyModel.v_0']

        """

    def incr(obj_or_cls, start=0):
        "Increment version of object and class or of class only"


class IIdentityInspector(Interface):

    def identify(obj_or_cls):
        """Return a string that can be used to uniquely identify ``obj_or_cls``
        :param obj_or_cls: object or class to identify
        """


class ICacheClient(Interface):

    def add(key, obj):
        pass

    def get(key):
        pass


class ICacheManager(Interface):
    pass


class ISerializer(Interface):
    def dumps(obj):
        """Return a flat binary representation of the object."""

    def loads(data):
        """Returns an instance of the object stored in the binary data."""


class ISerializationAdapter(Interface):

    name = Attribute('unique name registered for the serializer')

    def serialize(obj):
        pass

    def deserialize(data):
        pass


class ICacheHit(Interface):
    cache_key = Attribute("The cache key object")
    request = Attribute("The request object")


class ICacheMiss(Interface):
    cache_key = Attribute("The cache key object")
    request = Attribute("The request object")



import unittest

import mock
from nose_parameterized import parameterized

from pyramid_caching.ext.redis import RedisVersionWrapper
from pyramid_caching.exc import (
    VersionGetError,
    VersionMasterVersionError,
    VersionIncrementError,
    CacheDisabled,
)

from redis import (StrictRedis, RedisError)


def get_redis():
    return RedisVersionWrapper(StrictRedis())


KEY_VERSIONERS = [
    ("redis", get_redis, '0'),
]


class TestKeyVersioner(unittest.TestCase):

    TEST_VALUE = '4'  # Best creative testing value

    def setUp(self):
        get_redis().flush_all()

    @parameterized.expand(KEY_VERSIONERS)
    def test_interface(self, name, get_key_versioner, default_value):
        from zope.interface.verify import verifyObject
        from pyramid_caching.interfaces import IKeyVersioner

        key_versioner = get_key_versioner()

        self.assertTrue(verifyObject(IKeyVersioner, key_versioner))

    @parameterized.expand(KEY_VERSIONERS)
    @mock.patch('pyramid_caching.ext.redis.time')
    def test_get_multi(self, name, get_key_versioner, default_value, m_time):
        m_time.time.return_value = 1234.123
        key_versioner = get_key_versioner()

        KEYS = ['FOO', 'BAR', '2000']
        VERSIONS = [('cache', '1234'),
                    ('FOO', '0'),
                    ('BAR', '0'),
                    ('2000', '0')]

        self.assertEqual(key_versioner.get_multi(KEYS), VERSIONS)
        self.assertEqual(key_versioner.get_multi(KEYS), VERSIONS)

        for _ in range(int(self.TEST_VALUE)):
            key_versioner.incr('BAR')

        VERSIONS[2] = ('BAR', str(self.TEST_VALUE))

        self.assertEqual(key_versioner.get_multi(KEYS), VERSIONS)
        self.assertEqual(key_versioner.get_multi(KEYS), VERSIONS)

        for _ in range(42):
            key_versioner.incr('2000')

        VERSIONS[3] = ('2000', '42')

        self.assertEqual(key_versioner.get_multi(KEYS), VERSIONS)
        self.assertEqual(key_versioner.get_multi(KEYS), VERSIONS)


class TestRedisVersionClient(unittest.TestCase):
    """Test specifics of Redis client"""

    def setUp(self):
        self.redis_client = mock.Mock(name='RedisClient')
        self.version_store = RedisVersionWrapper(self.redis_client)

    def test_get_master_version(self):
        master_version = self.version_store._get_master_version()

        self.assertEqual(master_version, self.redis_client.get.return_value)
        self.redis_client.get.assert_called_once_with('cache')

    def test_set_master_version(self):
        with mock.patch('pyramid_caching.ext.redis.time') as m_t:
            m_t.time.return_value = 42
            self.version_store._set_master_version()

        self.redis_client.set.assert_called_once_with('cache', 42, nx=True)

    def test_get_multi(self):
        keys = ['one', 'two']
        self.redis_client.mget.return_value = ['42', '1', None]

        result = self.version_store.get_multi(keys)

        self.assertEqual(result, [('cache', '42'), ('one', '1'), ('two', '0')])
        self.redis_client.mget.assert_called_once_with(['cache', 'one', 'two'])

    def test_get_multi_no_master_version(self):
        keys = ['one', 'two']
        self.redis_client.mget.return_value = [None, '1', '2']
        self.redis_client.get.return_value = '84'

        with mock.patch('pyramid_caching.ext.redis.time') as m_t:
            m_t.time.return_value = 42
            result = self.version_store.get_multi(keys)

        expected = [('cache', '84'),
                    ('one', '1'),
                    ('two', '2')]
        self.assertEqual(result, expected)
        self.redis_client.set.assert_called_once_with('cache', 42, nx=True)

    def test_get_multi_no_master_version_after_reset_attempt(self):
        keys = []
        self.redis_client.mget.return_value = [None]
        self.redis_client.get.return_value = None

        with mock.patch('pyramid_caching.ext.redis.time') as m_t:
            m_t.time.return_value = 42
            with self.assertRaises(VersionMasterVersionError):
                self.version_store.get_multi(keys)

    def test_get_multi_redis_error_1(self):
        self.redis_client.mget.side_effect = RedisError()

        with self.assertRaises(VersionGetError):
            self.version_store.get_multi([])

    def test_get_multi_redis_error_2(self):
        self.redis_client.mget.return_value = [None]
        self.redis_client.set.side_effect = RedisError()

        with self.assertRaises(VersionMasterVersionError):
            self.version_store.get_multi([])

    def test_get_multi_redis_error_3(self):
        self.redis_client.mget.return_value = [None]
        self.redis_client.get.side_effect = RedisError()

        with self.assertRaises(VersionMasterVersionError):
            self.version_store.get_multi([])

    def test_incr(self):
        self.version_store.incr('FOO')
        self.redis_client.incr.assert_called_once_with('FOO')

    def test_incr_redis_error(self):
        self.redis_client.incr.side_effect = RedisError()

        with self.assertRaises(VersionIncrementError):
            self.version_store.incr('FOO')

    def test_inhibit_caching(self):
        self.redis_client.mget.return_value = ['off']

        with self.assertRaises(CacheDisabled):
            self.version_store.get_multi([])



import os
import unittest

import mock
from nose_parameterized import parameterized


class TestInclude(unittest.TestCase):

    def tearDown(self):
        if 'CACHING_ENABLED' in os.environ:
            del os.environ['CACHING_ENABLED']

    def test_include(self):
        from pyramid_caching import includeme

        config = mock.Mock(name='config')
        config.registry.settings = {'caching.enabled': True}

        with mock.patch('pyramid_caching.parse_settings') as m_parse:
            includeme(config)

        m_parse.assert_called_once_with(config)
        config.include.assert_has_calls([
            mock.call('.versioner'),
            mock.call('.key_versioner'),
            mock.call('.serializers'),
            mock.call('.cache'),
            ])

    @parameterized.expand([
        ('default', {}, {}, True),
        ('env_true', {'CACHING_ENABLED': 'true'}, {}, True),
        ('env_false', {'CACHING_ENABLED': 'false'}, {}, False),
        ('settings_true', {}, {'caching.enabled': 'true'}, True),
        ('settings_false', {}, {'caching.enabled': 'false'}, False),
        ('both', {'CACHING_ENABLED': 'true'}, {'caching.enabled': 'false'}, True),
        ('both2', {'CACHING_ENABLED': 'false'}, {'caching.enabled': 'true'}, False),
        ])
    def test_parse_settings(self, name, envs, settings, expected):
        from pyramid_caching import parse_settings

        config = mock.Mock(name='config')
        config.registry.settings = settings

        os.environ.update(envs)

        parse_settings(config)

        self.assertEqual(config.registry.settings['caching.enabled'], expected)



import unittest

from pyramid_caching.cache import CacheKey


class TestMetricsExtension(unittest.TestCase):
    def test_cache_hit(self):
        from pyramid_caching.ext.metrics import cache_hit
        event = DummyEvent()
        cache_hit(event)
        self.assertEqual(len(event.request.metrics.keys), 1)
        self.assertEqual(event.request.metrics.keys[0], ('cache.hit', 'a:b'))

    def test_cache_miss(self):
        from pyramid_caching.ext.metrics import cache_miss
        event = DummyEvent()
        cache_miss(event)
        self.assertEqual(len(event.request.metrics.keys), 1)
        self.assertEqual(event.request.metrics.keys[0], ('cache.miss', 'a:b'))


class DummyEvent:
    def __init__(self):
        self.request = DummyRequest()
        self.cache_key = CacheKey(['a', 'b'], [])


class DummyRequest:
    def __init__(self):
        self.metrics = DummyMetrics()


class DummyMetrics:
    def __init__(self):
        self.keys = []

    def incr(self, key):
        self.keys.append(key)



import pickle
import unittest

from pyramid import testing

from pyramid_caching.serializers import (
    PICKLE_PROTOCOL,
    SerializerUtility,
    SERIALIZER_META_VERSION,
    )


class DummyRegistry(object):
    def queryAdapter(self, object, interface, name=None, default=None):
        # Utility should not query registry when passed an adapter.
        raise AssertionError()


class DummyAdapter(object):
    name = "dummy"

    def serialize(self, obj):
        return obj.upper()

    def deserialize(self, data):
        return data.lower()


class TestSerializers(unittest.TestCase):
    def setUp(self):
        self.config = testing.setUp()

    def tearDown(self):
        testing.tearDown()

    def test_implements_interface(self):
        from zope.interface.verify import verifyClass
        from pyramid_caching.interfaces import ISerializer
        verifyClass(ISerializer, SerializerUtility)

    def test_encode_serializer_type(self):
        utility = SerializerUtility(DummyRegistry())
        data = utility.dumps("object", adapter=DummyAdapter())
        data = pickle.loads(data)
        self.assertEqual(data['type'], "dummy")

    def test_encode_meta_format_version(self):
        utility = SerializerUtility(DummyRegistry())
        data = utility.dumps("object", adapter=DummyAdapter())
        data = pickle.loads(data)
        self.assertEqual(data['version'], SERIALIZER_META_VERSION)

    def _create_data(self,
                     meta_type=DummyAdapter.name,
                     meta_version=SERIALIZER_META_VERSION,
                     payload="OBJECT"):
        data = {
            'type': meta_type,
            'version': meta_version,
            'payload': payload,
            }
        return pickle.dumps(data, protocol=PICKLE_PROTOCOL)

    def test_invalid_meta_format_version(self):
        utility = SerializerUtility(DummyRegistry())
        data = self._create_data(meta_version=9999)
        self.assertIsNone(utility.loads(data))

    def test_encode_payload(self):
        utility = SerializerUtility(DummyRegistry())
        data = utility.dumps("object", adapter=DummyAdapter())
        data = pickle.loads(data)
        self.assertEqual(data['payload'], "OBJECT")

    def test_register_serializer(self):
        utility = SerializerUtility(self.config.registry)
        utility.register_serialization_adapter(str, DummyAdapter)
        data = utility.dumps("object")
        data = pickle.loads(data)
        self.assertEqual(data['type'], "dummy")

    def test_query_serializer(self):
        utility = SerializerUtility(self.config.registry)
        utility.register_serialization_adapter(str, DummyAdapter)
        data = utility.dumps("object")
        data = pickle.loads(data)
        self.assertEqual(data['payload'], "OBJECT")

    def test_query_deserializer(self):
        utility = SerializerUtility(self.config.registry)
        utility.register_serialization_adapter(str, DummyAdapter)
        data = self._create_data()
        obj = utility.loads(data)
        self.assertEqual(obj, "object")

    def test_missing_deserializer(self):
        from pyramid_caching.serializers import DeserializationError
        utility = SerializerUtility(self.config.registry)
        data = {
            'type': 'alien',
            'version': SERIALIZER_META_VERSION,
            'payload': "oBjEcT",
            }
        data = pickle.dumps(data, protocol=PICKLE_PROTOCOL)
        with self.assertRaises(DeserializationError):
            utility.loads(data)

    def test_missing_serializer(self):
        from pyramid_caching.serializers import SerializationError
        utility = SerializerUtility(self.config.registry)

        class UnknownType:
            pass

        with self.assertRaises(SerializationError):
            utility.dumps(UnknownType())



import unittest

import mock
from nose_parameterized import parameterized

from pyramid_caching.ext.redis import RedisCacheWrapper
from pyramid_caching.exc import (
    CacheKeyAlreadyExists,
    CacheAddError,
    CacheGetError,
    )

from redis import RedisError


class MockClient(object):
    def __init__(self, m_set, m_get, m_flush):
        self.m_set = m_set
        self.m_get = m_get
        self.m_flush = m_flush

    def set(self, *args, **kwargs):
        return self.m_set(*args, **kwargs)

    def get(self, *args, **kwargs):
        return self.m_get(*args, **kwargs)

    def flushall(self, *args, **kwargs):
        return self.m_flush(*args, **kwargs)


def get_redis(**kwargs):
    m_set = mock.Mock(name='add')
    m_get = mock.Mock(name='get')
    m_flush = mock.Mock(name='get')
    m_client = MockClient(m_set=m_set, m_get=m_get, m_flush=m_flush)
    return (m_set, m_get, m_flush, RedisCacheWrapper(m_client))

CACHE_CLIENTS = [
    ("redis", get_redis, '0'),
]


class TestCacheClient(unittest.TestCase):

    def setUp(self):
        m_set, m_get, m_flush, cache = get_redis()
        cache.flush_all()

    @parameterized.expand(CACHE_CLIENTS)
    def test_interface(self, name, get_cache, default_value):
        from zope.interface.verify import verifyObject
        from pyramid_caching.interfaces import ICacheClient

        m_set, m_get, m_flush, cache = get_cache()

        self.assertTrue(verifyObject(ICacheClient, cache))

    @parameterized.expand(CACHE_CLIENTS)
    def test_flush_all(self, name, get_cache, default_value):
        """Note: we use flush_all() for testing, so let's validate it first"""
        m_set, m_get, m_flush, cache = get_cache()

        cache.flush_all()
        m_flush.assert_called_once()

    @parameterized.expand(CACHE_CLIENTS)
    def test_add_ok(self, name, get_cache, default_value):
        m_set, m_get, m_flush, cache = get_cache()

        ex = 3600 * 24 * 7
        cache.add('FOO', 'BAR', expiration=ex)
        m_set.assert_called_once_with('FOO', 'BAR', ex=ex, nx=True)

    @parameterized.expand(CACHE_CLIENTS)
    def test_add_with_expiration(self, name, get_cache, default_value):
        m_set, m_get, m_flush, cache = get_cache()

        ex = 42
        cache.add('FOO', 'BAR', expiration=ex)
        m_set.assert_called_once_with('FOO', 'BAR', ex=ex, nx=True)

    @parameterized.expand(CACHE_CLIENTS)
    def test_add_exists(self, name, get_cache, default_value):
        m_set, m_get, m_flush, cache = get_cache()

        m_set.return_value = None
        with self.assertRaises(CacheKeyAlreadyExists):
            cache.add('FOO', 'BAR')

    @parameterized.expand(CACHE_CLIENTS)
    def test_add_network_error(self, name, get_cache, default_value):
        m_set, m_get, m_flush, cache = get_cache()

        m_set.side_effect = RedisError()
        with self.assertRaises(CacheAddError):
            cache.add('FOO', 'BAR')

    @parameterized.expand(CACHE_CLIENTS)
    def test_get_ok(self, name, get_cache, default_value):
        m_set, m_get, m_flush, cache = get_cache()

        m_get.return_value = 'BAR'
        self.assertEquals(cache.get('FOO'), 'BAR')

    @parameterized.expand(CACHE_CLIENTS)
    def test_get_network_error(self, name, get_cache, default_value):
        m_set, m_get, m_flush, cache = get_cache()

        m_get.side_effect = RedisError()

        with self.assertRaises(CacheGetError):
            cache.get('FOO')



from pyramid.view import view_config
from pyramid.httpexceptions import HTTPNotFound

from pyramid_caching.cache import cache_factory

from example.model import User


def includeme(config):
    config.add_route(name='user', pattern='/users/{user_id}')
    config.scan(__name__)


@view_config(
    route_name='user',
    renderer='json',
    request_method='GET',
    decorator=cache_factory(depends_on={User: {'matchdict': ['user_id']}}),
    )
def get_user(request):
    user = User.get(request.matchdict['user_id'])
    if user is None:
        raise HTTPNotFound()
    return {
        'id': user.id,
        'name': user.name,
    }



from __future__ import absolute_import

import os
import time

from redis import StrictRedis, RedisError

from pyramid_caching.exc import (
    CacheAddError,
    CacheGetError,
    CacheKeyAlreadyExists,
    VersionGetError,
    VersionIncrementError,
    VersionMasterVersionError,
    CacheDisabled,
)


def includeme(config):
    if not config.registry.settings['caching.enabled']:
        return
    include_cache_store(config)
    include_version_store(config)


def include_cache_store(config):
    uri = os.environ['CACHE_STORE_REDIS_URI']
    client = StrictRedis.from_url(uri)
    config.add_cache_client(RedisCacheWrapper(client))


def include_version_store(config):
    uri = os.environ['VERSION_STORE_REDIS_URI']
    client = StrictRedis.from_url(uri)
    config.add_key_version_client(RedisVersionWrapper(client))


class RedisCacheWrapper(object):

    def __init__(self, client):
        self.default_expiration = 3600 * 24 * 7  # 7 days
        self.client = client

    def add(self, key, value, expiration=None):
        """Create a cache entry. Raise CacheKeyAlreadyExists is this entry
        already exists.

        Note: Redis will only LRU evince volatile keys. (Default: 7 days)
        """
        if expiration is None:
            expiration = self.default_expiration

        try:
            rvalue = self.client.set(key, value, ex=expiration, nx=True)
        except RedisError as error:
            raise CacheAddError(error)

        if rvalue is None:
            raise CacheKeyAlreadyExists(key)

    def get(self, key):
        try:
            return self.client.get(key)
        except RedisError as error:
            raise CacheGetError(error)

    def flush_all(self):
        self.client.flushall()


class RedisVersionWrapper(object):
    """Redis implementation of the IKeyVersioner interface.

    This KeyVersioner use Redis ability to namespace content with seperate
    db, thus not prefixing the keys with anything like 'version:'.

    The default value for this implementation is 0 since the INCR operation
    enforce a value of 1 after executing on a non-existing key.
    See `Redis documentation for INCR <http://redis.io/commands/incr>`_

    Note: the return value are always string type.

    Notes about the Consistency and master-version:

    To ensure cache data consistency the model versioning must be consistent,
    thus a single failed model version increment would result in stale cache.

    To achieve consistent model versioning we must protect against two
    situations:

    - Glitch / temporary unavailability

    The model version increment operation raises VersionIncrementError
    exception is raised. In this case the write operation must be rollback.

    - Permanent loss of the version-store (and recovery with an empty one)

    The master-version is used together with dependency versions. If missing,
    this master-version is initialized to a new value. When a new empty
    version-store is reacheable, the master-version ends up being different
    than the previous one, hence effectively invalidating the whole cache.

    Note: the special master-version 'off' will inhibit caching while still
    maintaining the model versions.
    """

    MASTER_VERSION_KEY = 'cache'
    MASTER_VERSION_DISABLE_VALUE = 'off'

    def __init__(self, client):
        self.client = client

    def _get_master_version(self):
        """Return the master-version or None if the key is missing"""
        return self.client.get(self.MASTER_VERSION_KEY)

    def _generate_master_version(self):
        return int(time.time())

    def _set_master_version(self):
        """The set operation will silently fail if the key already exists.
        Thus we don't know what is the value of the stored master-version"""

        self.client.set(self.MASTER_VERSION_KEY,
                        self._generate_master_version(),
                        nx=True)

    def _handle_master_version(self, versions):
        if versions[0] is None:
            try:
                self._set_master_version()
                versions[0] = self._get_master_version()
            except RedisError as error:
                raise VersionMasterVersionError(error)

        if versions[0] is None:
            raise VersionMasterVersionError(
                "Still no master version after reset attempt")

        if versions[0] == self.MASTER_VERSION_DISABLE_VALUE:
            raise CacheDisabled('Disabled by master_version')

    def get_multi(self, keys):
        """Return an ordered list of tuple (key, value). The value default to 0
        """
        keys_with_master = [self.MASTER_VERSION_KEY] + keys

        try:
            versions = self.client.mget(keys_with_master)
        except RedisError as error:
            raise VersionGetError(error)

        self._handle_master_version(versions)

        versions = [v if v is not None else '0' for v in versions]

        return zip(keys_with_master, versions)

    def incr(self, key):
        """Increment a version. If the key was missing, the new value is 1"""
        try:
            self.client.incr(key)
        except RedisError as error:
            raise VersionIncrementError(error)

    def flush_all(self):
        self.client.flushall()




class Base(Exception):
    """Base exception for pyramid_caching"""


class CacheDisabled(Base):
    """Cache is disabled """


class CacheError(Base):
    """Base exception for cache client"""


class CacheKeyAlreadyExists(CacheError):
    """Trying to create an existing key in cache"""


class CacheAddError(CacheError):
    """Error on cache key creation"""


class CacheGetError(CacheError):
    """Error on cache key retrieval"""


class VersionError(Base):
    """Base exception for version client"""


class VersionGetError(VersionError):
    """Error on model version increment operation"""


class VersionIncrementError(VersionError):
    """Error on model version increment operation"""


class VersionMasterVersionError(VersionError):
    """Error on master version operation"""


class SerializationError(Exception):
    """An error that occurs if no encoder can be found matching the object
    type to serialize."""


class DeserializationError(Exception):
    """An error that occurs when valid cached data could not be decoded."""



from sqlalchemy import Column, Integer, String, MetaData, ForeignKey, Text
from sqlalchemy.engine import engine_from_config
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker, relationship

metadata = MetaData()

Session = sessionmaker()


def includeme(config):
    config.include('pyramid_caching.ext.sqlalchemy')
    engine = engine_from_config(config.registry.settings)
    metadata.create_all(engine)
    Session.configure(bind=engine)

    config.register_sqla_session_caching_hook(Session)
    config.register_sqla_base_class(Base)

    config.scan(__name__)


class Base(declarative_base(metadata=metadata)):
    __abstract__ = True

    @classmethod
    def get(cls, ids):
        return Session().query(cls).get(ids)


class User(Base):

    __tablename__ = 'user'

    id = Column('id', Integer, primary_key=True)

    name = Column('name', String(32))

    notes = relationship('UserNote', backref='user',
                         cascade='all, delete-orphan')

    def __repr__(self):
        return '<User id=%s>' % self.id


class UserNote(Base):

    __tablename__ = 'user_note'

    id = Column('id', Integer, primary_key=True, autoincrement=True)

    user_id = Column('user_id', ForeignKey('user.id', ondelete='cascade'),
                     primary_key=True)

    content = Column('name', Text())

    def __repr__(self):
        return '<UserNote id=%2s user_id=%s>' % (self.id, self.user_id)



import unittest


class ViewCacheHitEventTests(unittest.TestCase):
    def test_class_implements_interface(self):
        from pyramid_caching.events import ViewCacheHit
        from pyramid_caching.interfaces import ICacheHit
        from zope.interface.verify import verifyClass
        verifyClass(ICacheHit, ViewCacheHit)

    def test_instance_implements_interface(self):
        from pyramid_caching.events import ViewCacheHit
        from pyramid_caching.interfaces import ICacheHit
        from zope.interface.verify import verifyObject
        request = DummyRequest()
        event = ViewCacheHit(None, request)
        verifyObject(ICacheHit, event)

    def test_attributes(self):
        from pyramid_caching.events import ViewCacheHit
        key = object()
        request = DummyRequest()
        event = ViewCacheHit(key, request)
        self.assertEqual(event.request, request)
        self.assertEqual(event.cache_key, key)


class ViewCacheMissEventTests(unittest.TestCase):
    def test_class_implements_interface(self):
        from pyramid_caching.events import ViewCacheMiss
        from pyramid_caching.interfaces import ICacheMiss
        from zope.interface.verify import verifyClass
        verifyClass(ICacheMiss, ViewCacheMiss)

    def test_instance_implements_interface(self):
        from pyramid_caching.events import ViewCacheMiss
        from pyramid_caching.interfaces import ICacheMiss
        from zope.interface.verify import verifyObject
        request = DummyRequest()
        event = ViewCacheMiss("a_key", request)
        verifyObject(ICacheMiss, event)

    def test_attributes(self):
        from pyramid_caching.events import ViewCacheMiss
        key = object()
        request = DummyRequest()
        event = ViewCacheMiss(key, request)
        self.assertEqual(event.request, request)
        self.assertEqual(event.cache_key, key)


class DummyRequest:
    pass



import os
import unittest

import webtest

from pyramid.config import Configurator
from pyramid.decorator import reify


def setupPackage():
    os.environ['CACHE_STORE_REDIS_URI'] = 'redis://127.0.0.1:6379/5'
    os.environ['VERSION_STORE_REDIS_URI'] = 'redis://127.0.0.1:6379/8'


def tearDownPackage():
    os.environ.pop('CACHE_STORE_REDIS_URI', None)
    os.environ.pop('VERSION_STORE_REDIS_URI', None)


class Base(unittest.TestCase):

    maxDiff = None

    @reify
    def config(self):
        _config = Configurator(settings={
            'sqlalchemy.url': 'sqlite:///:memory:',
            })
        _config.include('example')
        self.addCleanup(delattr, self, 'config')

        def flush_cache():
            _config.get_cache_client().flush_all()

        if hasattr(_config, 'get_cache_client'):
            self.addCleanup(flush_cache)
        return _config

    @reify
    def app(self):
        self.addCleanup(delattr, self, 'app')
        return webtest.TestApp(self.config.make_wsgi_app())

    @property
    def registry(self):  # pragma no cover
        return self.config.registry

    @property
    def versioner(self):
        return self.config.get_versioner()

    @property
    def key_versioner_client(self):
        return self.config.get_key_version_client()

    @property
    def cache_client(self):
        return self.config.get_cache_client()

    @property
    def cache_manager(self):
        return self.config.get_cache_manager()



import inspect
import unittest

from nose_parameterized import parameterized
from zope.interface import implementer

from pyramid_caching.interfaces import IIdentityInspector
from pyramid_caching.versioner import Versioner


class BasicModel(object):
    def __init__(self, id):
        self.id = id

    def __hash__(self):
        return self.id


@implementer(IIdentityInspector)
class BasicModelIdentityInspector(object):

    def _get_cls_identity(self, cls):
        return '%s.%s' % (cls.__module__, cls.__name__)

    def _get_obj_identity(self, obj):
        return '%s.id_%s' % (self._get_cls_identity(obj.__class__), hash(obj))

    def identify(self, obj_or_cls):
        if inspect.isclass(obj_or_cls):
            identity = self._get_cls_identity(obj_or_cls)
        else:
            identity = self._get_obj_identity(obj_or_cls)

        return identity


def get_basic():
    def instantiate_model(id=4):
        return BasicModel(id)

    class TestKeyVersioner(object):
        def __init__(self):
            self._d = dict()

        def get_multi(self, keys):
            return [(k, self._d.setdefault(k, 0)) for k in keys]

        def incr(self, key):
            self._d[key] += 1
    key_versioner = TestKeyVersioner()

    id_inspector = BasicModelIdentityInspector()

    versioner = Versioner(key_versioner, id_inspector.identify)
    return versioner, instantiate_model, BasicModel


class TestBasic(unittest.TestCase):

    @parameterized.expand([
        ("with_basic_impl", get_basic),
    ])
    def test_get_multi_keys(self, test_name, test_fixture_func):
        result = test_fixture_func()
        versioner, model_factory, model_cls = result

        model1 = model_factory(id=42)
        model2 = model_factory(id=43)
        model3 = model_factory(id=44)

        result = versioner.get_multi_keys(
            [model1, model2, model3, model_cls]
        )

        expected = [
            'pyramid_caching.tests.unittests.test_versioners.BasicModel.id_42:v=0',
            'pyramid_caching.tests.unittests.test_versioners.BasicModel.id_43:v=0',
            'pyramid_caching.tests.unittests.test_versioners.BasicModel.id_44:v=0',
            'pyramid_caching.tests.unittests.test_versioners.BasicModel:v=0',
        ]

        self.assertEqual(result, expected)

    @parameterized.expand([
        ("basic", get_basic),
    ])
    def test_incr(self, test_name, test_fixture_func):
        result = test_fixture_func()
        versioner, model_factory, model_cls = result

        model1 = model_factory()

        key_cls_v0 = versioner.get_multi_keys([model_cls])
        key_id1_v0 = versioner.get_multi_keys([model1])

        versioner.incr(model1)

        key_id1_v1 = versioner.get_multi_keys([model1])
        key_cls_v1 = versioner.get_multi_keys([model_cls])

        self.assertNotEqual(key_id1_v0, key_id1_v1)
        self.assertNotEqual(key_cls_v0, key_cls_v1)



from collections import namedtuple
import hashlib
import logging

from zope.interface import implementer, classImplements

from pyramid_caching.events import ViewCacheHit, ViewCacheMiss
from pyramid_caching.interfaces import (
    ICacheClient,
    ICacheManager,
    )
from pyramid_caching.exc import (
    CacheDisabled,
    )

log = logging.getLogger(__name__)


def includeme(config):
    """Set up a request method to access the cache manager utility."""
    config.add_directive('get_cache_client', get_cache_client,
                         action_wrap=False)

    config.add_directive('get_cache_manager', get_cache_manager,
                         action_wrap=False)

    config.add_directive('add_cache_client', add_cache_client)

    config.add_request_method(get_cache_manager, 'cache_manager', reify=True)

    def register():
        cache_client = config.get_cache_client()
        versioner = config.get_versioner()
        serializer = config.get_serializer()
        manager = Manager(config.registry, versioner, cache_client, serializer)
        config.registry.registerUtility(manager)
        log.debug('registering cache manager %r', manager)

    config.action((__name__, 'cache_manager'), register, order=2)


def get_cache_client(config_or_request):
    return config_or_request.registry.getUtility(ICacheClient)


def get_cache_manager(config_or_request):
    return config_or_request.registry.getUtility(ICacheManager)


def add_cache_client(config, client):
    if not ICacheClient.implementedBy(client.__class__):
        log.debug('assuming %r implements %r', client.__class__, ICacheClient)
        classImplements(client.__class__, ICacheClient)

    def register():
        log.debug('registering cache client %r', client)
        config.registry.registerUtility(client, ICacheClient)

    config.action((__name__, 'cache_client'), register, order=0)


@implementer(ICacheManager)
class Manager(object):

    def __init__(self, registry, versioner, cache_client, serializer):
        self.versioner = versioner
        self.cache_client = cache_client
        self.serializer = serializer
        self.registry = registry

    def get_or_cache(self, get_result, prefixes, dependencies):
        versioned_keys = self.versioner.get_multi_keys(dependencies)
        key = CacheKey(prefixes, versioned_keys)

        cache_content = self.cache_client.get(str(key))

        if cache_content is not None:
            result = self.serializer.loads(cache_content)
            return CacheResult.hit(key, result)
        else:
            result = get_result()
            data = self.serializer.dumps(result)
            self.cache_client.add(str(key), data)
            return CacheResult.miss(key, result)


class CacheKey(object):

    """A key identifying a version of a complex resource in a cache.

    This object specifies the formatting of cache keys from a list of base
    resource names (for example, a unique way to identify a Pyramid view
    callable) and a list of names that define the current context of the view
    (for example, a list of model names and the primary keys matching the
    route).

    ::
       key = CacheKey(['mypackage.views', 'hello_view'], ['user:bob'])
       key.root() --> 'mypackage.views:hello_view'
       str(key) --> 'mypackage.views:hello_view:user:bob'

    """

    def __init__(self, bases, dependencies):
        self.bases = bases
        self.dependencies = dependencies

    def root(self):
        """The static part that refers to a single view or model class."""
        return ':'.join(self.bases)

    def key(self):
        """The unique cache key identifying a resource and its context."""
        return self.root() + ':' + ':'.join(self.dependencies)

    def __str__(self):
        return self.key()


_CacheResultInfo = namedtuple("CacheResultInfo", ["key", "hit"])


class CacheResult(object):

    """A versioned cache response based on a unique key."""

    def __init__(self, cache_key, data, hit):
        self._cache_key = cache_key
        self.data = data
        self._hit = hit

    @classmethod
    def hit(cls, cache_key, data):
        """Specify that this data was fetched from the cache."""
        return cls(cache_key, data, True)

    @classmethod
    def miss(cls, cache_key, data):
        """Specify that this data was loaded from the application."""
        return cls(cache_key, data, False)

    def key_hash(self):
        """Unique hash to identify this result in the cache.

        A string corresponding to a cryptographic hash digest of the cache key
        that uniquely defines the version of this result.
        """
        m = hashlib.sha1()
        m.update(str(self._cache_key))
        return m.hexdigest()

    def info(self):
        return _CacheResultInfo(self._cache_key, self._hit)


class cache_factory(object):

    """Decorator that defines the model dependencies for a view method.

    ::
       @cache_factory(depends_on={User: {'matchdict': 'user'}})
       def hello_view(context, request):
           user = User(request.matchdict['user'])
           return "Hello, {}".format(user.name)

    """

    def __init__(self, depends_on=None):
        self.depends_on = depends_on

    def __call__(self, view):
        return ViewCacheDecorator(view, self.depends_on)


class ViewCacheDecorator(object):

    def __init__(self, view, depends_on=None):
        self.view = view
        self.depends_on = depends_on

    def __call__(self, context, request):
        def get_result():
            return self.view(context, request)

        def nocache_result():
            response = get_result()
            response.headers['X-View-Cache'] = 'DISABLED'
            return response

        if not request.registry.settings['caching.enabled']:
            return nocache_result()

        cache_manager = request.cache_manager

        prefixes = [self.view.__module__, self.view.__name__]
        dependencies = self.get_dependencies(request.matchdict)

        try:
            result = cache_manager.get_or_cache(get_result,
                                                prefixes,
                                                dependencies)
        except CacheDisabled:
            return nocache_result()

        response = result.data
        result_info = result.info()
        if result_info.hit:
            request.registry.notify(ViewCacheHit(result_info.key, request))
            response.headers['X-View-Cache'] = 'HIT'
        else:
            request.registry.notify(ViewCacheMiss(result_info.key, request))
            response.headers['X-View-Cache'] = 'MISS'
        response.headers['ETag'] = result.key_hash()
        return response

    def get_dependencies(self, matchdict):
        dependencies = []
        for cls, options in self.depends_on.iteritems():

            ids_dict = {}
            if 'matchdict' in options:
                for match in options['matchdict']:
                    ids_dict.update({match: matchdict[match]})

            dependencies.append((cls, ids_dict))
        return dependencies



"""An extension that records the performance of the caching layer by sending
information about cache hits and cache misses events to a statistics aggregator
via pyramid_metrics.
"""

from pyramid.events import subscriber

from pyramid_caching.events import ViewCacheHit, ViewCacheMiss


def includeme(config):
    """Include subscriptions to cache hit and cache miss events into the
    specified Pyramid configurator. This will allow the application to
    automatically provide cache performance statistics.

    To activate this extension, call the configurator method:

    .. code-block:: python

       config.include('pyramid_caching.ext.metrics')

    """
    config.include('pyramid_metrics')
    config.scan()


@subscriber(ViewCacheHit)
def cache_hit(event):
    count_view_cache_event(event, 'hit')


@subscriber(ViewCacheMiss)
def cache_miss(event):
    count_view_cache_event(event, 'miss')


def count_view_cache_event(event, access):
    metrics = event.request.metrics
    metrics.incr(('cache.%s' % access, event.cache_key.root()))



import os
from pyramid.settings import asbool


def includeme(config):
    """Pyramid_caching main includeme.

    The extension is enabled by default. You can disable all mechanisms with:

    - environment variable CACHING_ENABLED=false
    - setting caching.enabled = false
    """
    parse_settings(config)

    if config.registry.settings['caching.enabled']:
        config.include('.versioner')
        config.include('.key_versioner')
        config.include('.serializers')
        config.include('.cache')


def parse_settings(config):
    cache_enabled = config.registry.settings.get('caching.enabled', 'true')
    cache_enabled = os.getenv('CACHING_ENABLED', cache_enabled)
    config.registry.settings['caching.enabled'] = asbool(cache_enabled)



from __future__ import absolute_import

import inspect
import logging

from zope.interface import implementer

from pyramid_caching.interfaces import IIdentityInspector

from sqlalchemy import event

log = logging.getLogger(__name__)


def includeme(config):
    config.add_directive('register_sqla_session_caching_hook',
                         register_sqla_session_caching_hook)

    config.add_directive('register_sqla_base_class',
                         register_sqla_base_class)

    if config.registry.settings['caching.enabled']:
        identity_inspector = SqlAlchemyIdentityInspector()

        config.registry.registerUtility(identity_inspector,
                                        provided=IIdentityInspector)


def register_sqla_base_class(config, base_cls):
    if not config.registry.settings['caching.enabled']:
        return

    registry = config.registry

    identity_inspector = registry.getUtility(IIdentityInspector)

    def identify(model):
        return identity_inspector.identify(model)

    registry.registerAdapter(identify, required=[base_cls],
                             provided=IIdentityInspector)

    registry.registerAdapter(identify, required=[base_cls.__class__],
                             provided=IIdentityInspector)


def register_sqla_session_caching_hook(config, session_cls):
    if not config.registry.settings['caching.enabled']:
        return

    def register():
        versioner = config.get_versioner()
        registry = config.registry

        def identify(entity):
            return registry.queryAdapter(entity, IIdentityInspector)

        def on_before_commit(session):
            identities = []

            for entity in session.dirty:
                identities.append(identify(entity))

            for entity in session.deleted:
                identities.append(identify(entity))

            def after_commit(session):
                for identity in identities:
                    versioner.incr(identity)

            event.listen(session, 'after_commit', after_commit)

        event.listen(session_cls, 'before_commit', on_before_commit)

    config.action((__name__, 'session_caching_hook'), register, order=3)


@implementer(IIdentityInspector)
class SqlAlchemyIdentityInspector(object):

    def identify(self, obj_or_cls):
        tablename = obj_or_cls.__tablename__

        if inspect.isclass(obj_or_cls):
            return tablename

        ids = ''

        # with a table user_message with a composite primary key user_id and id
        # an object user_message(user_id=123, id=456) will give:
        # 'user_message:user_id=123:id=456'

        # TODO: if table has no primary keys :-/
        table = obj_or_cls.__table__

        ids += ':'.join(['%s=%s' % (col_name, getattr(obj_or_cls, col_name))
                         for col_name in table.primary_key.columns.keys()])

        return ':'.join([tablename, ids])



import inspect
import logging

from zope.interface import implementer

from pyramid_caching.interfaces import (
    IIdentityInspector,
    IVersioner,
    )

log = logging.getLogger(__name__)


def includeme(config):
    registry = config.registry

    def identify(model_obj_or_cls):
        return registry.queryAdapter(model_obj_or_cls, IIdentityInspector)

    registry.registerAdapter(lambda x: x, required=[str],
                             provided=IIdentityInspector)

    registry.registerAdapter(lambda x: str(x), required=[unicode],
                             provided=IIdentityInspector)

    registry.registerAdapter(lambda x: str(x), required=[int],
                             provided=IIdentityInspector)

    registry.registerAdapter(lambda x: str(x), required=[float],
                             provided=IIdentityInspector)

    registry.registerAdapter(
        lambda x: TupleIdentityInspector(identify).identify(x),
        required=[tuple],
        provided=IIdentityInspector,
        )

    registry.registerAdapter(
        lambda x: DictIdentityInspector(identify).identify(x),
        required=[dict],
        provided=IIdentityInspector,
        )

    config.add_directive('get_versioner', get_versioner, action_wrap=False)
    config.add_request_method(get_versioner, 'versioner', reify=True)

    def register():
        key_versioner = config.get_key_version_client()
        versioner = Versioner(key_versioner, identify)
        config.registry.registerUtility(versioner)
        log.debug('registering versioner %r', versioner)

    config.action((__name__, 'versioner'), register, order=1)


def get_versioner(config_or_request):
    return config_or_request.registry.getUtility(IVersioner)


@implementer(IIdentityInspector)
class TupleIdentityInspector(object):

    def __init__(self, identify):
        self._identify = identify

    def identify(self, tuple_or_list):
        return ':'.join([self._identify(elem) for elem in tuple_or_list])


@implementer(IIdentityInspector)
class DictIdentityInspector(object):

    def __init__(self, identify):
        self._identify = identify

    def identify(self, dict_like):
        elems = ['%s=%s' % (self._identify(k), self._identify(v))
                 for k, v in dict_like.iteritems()]
        return ':'.join(elems)


@implementer(IVersioner)
class Versioner(object):

    def __init__(self, key_versioner, identify):
        self.key_versioner = key_versioner
        self.identify = identify

    def get_multi_keys(self, things):
        keys = [self.identify(anything) for anything in things]

        versiontuples = self.key_versioner.get_multi(keys)

        return ['%s:v=%s' % (key, version) for (key, version) in versiontuples]

    def incr(self, obj_or_cls, start=0):
        self.key_versioner.incr(self.identify(obj_or_cls))

        if not inspect.isclass(obj_or_cls):  # increment model class version
            identity = self.identify(obj_or_cls.__class__)
            self.key_versioner.incr(identity)



import unittest

from pyramid import testing
from pyramid.response import Response

from pyramid_caching.cache import CacheResult


class CacheManagerTests(unittest.TestCase):
    def setUp(self):
        self.config = testing.setUp()
        self.registry = self.config.registry

    def tearDown(self):
        testing.tearDown()

    def _make_one(self, cached_value):
        from pyramid_caching.cache import Manager
        return Manager(self.registry,
                       DummyVersioner(),
                       DummyClient(cached_value),
                       DummySerializer())

    def test_key_root_from_bases(self):
        manager = self._make_one(True)
        result = manager.get_or_cache(None, ['a', 'b'], [])
        self.assertEqual(result._cache_key.root(), 'a:b')

    def test_key_from_dependencies(self):
        manager = self._make_one(True)
        result = manager.get_or_cache(None, ['a', 'b'], ['c', 'd'])
        self.assertEqual(result._cache_key.key(), 'a:b:c:d')

    def test_cache_miss_when_client_returns_none(self):
        def get_result():
            pass

        manager = self._make_one(None)
        result = manager.get_or_cache(get_result, [], [])
        self.assertFalse(result.info().hit)

    def test_cache_hit_when_client_returns_object(self):
        def get_result():
            pass

        manager = self._make_one(object())
        result = manager.get_or_cache(get_result, ['a', 'b'], ['c', 'd'])
        self.assertTrue(result.info().hit)

    def test_cache_miss_loads_data_from_function(self):
        def get_result():
            return "loaded"

        manager = self._make_one(None)
        result = manager.get_or_cache(get_result, [], [])
        self.assertEqual(result.data, "loaded")


class ViewCacheDecoratorTests(unittest.TestCase):
    def setUp(self):
        self.config = testing.setUp(settings={
            'caching.enabled': False,
            })

    def tearDown(self):
        testing.tearDown()

    def _view(self, context, request):
        return Response('ok')

    def test_decorates_method(self):
        from pyramid_caching.cache import cache_factory

        @cache_factory()
        def view(context, request):
            return Response('ok')

        request = testing.DummyRequest()
        response = view(None, request)
        self.assertEqual(response.body, "ok")

    def _make_one(self, depends_on=None, hit=True):
        from pyramid_caching.cache import ViewCacheDecorator

        request = testing.DummyRequest()
        request.registry.settings['caching.enabled'] = True
        request.cache_manager = DummyCacheManager(hit)

        if depends_on is None:
            depends_on = {}
        return request, ViewCacheDecorator(self._view, depends_on=depends_on)

    def test_key_base_from_view_name(self):
        request, deco = self._make_one()
        deco(None, request)
        self.assertEqual(request.cache_manager.prefixes, [__name__, '_view'])

    def test_key_dependencies_from_request(self):
        request, deco = self._make_one(depends_on={
            'User': {
                'matchdict': ['user'],
                }
            })
        request.matchdict = {'user': '123'}
        deco(None, request)
        self.assertEqual(request.cache_manager.dependencies,
                         [('User', {'user': '123'})])

    def test_cache_miss_calls_view(self):
        request, deco = self._make_one()
        deco(None, request)
        self.assertEqual(request.cache_manager.get_result().body, "ok")

    def test_sets_etag_from_result(self):
        request, deco = self._make_one()
        response = deco(None, request)
        key_hash = 'a62f2225bf70bfaccbc7f1ef2a397836717377de'  # SHA1 of 'key'
        self.assertEqual(response.headers['ETag'], key_hash)

    def test_sets_cache_result_header(self):
        request, deco = self._make_one()
        response = deco(None, request)
        self.assertEqual(response.headers['X-View-Cache'], 'HIT')

    def test_cache_disabled_result_header(self):
        from pyramid_caching.cache import ViewCacheDecorator
        request = testing.DummyRequest()
        deco = ViewCacheDecorator(self._view, depends_on={})
        response = deco(None, request)
        self.assertEqual(response.headers['X-View-Cache'], 'DISABLED')

    def _register_event_listener(self, iface):
        events = []

        def listener(event):
            events.append(event)

        self.config.registry.registerHandler(listener, (iface,))
        return events

    def test_cache_hit_event(self):
        from pyramid_caching.interfaces import ICacheHit

        hit_events = self._register_event_listener(ICacheHit)
        request, deco = self._make_one()
        deco(None, request)

        self.assertEqual(len(hit_events), 1)

    def test_cache_miss_event(self):
        from pyramid_caching.interfaces import ICacheMiss

        miss_events = self._register_event_listener(ICacheMiss)
        request, deco = self._make_one(hit=False)
        deco(None, request)
        self.assertEqual(len(miss_events), 1)


class DummyVersioner:
    def get_multi_keys(self, dependencies):
        return dependencies


class DummyClient:
    def __init__(self, cached_value):
        self._cached_value = cached_value

    def get(self, key):
        return self._cached_value

    def add(self, key, value):
        pass


class DummySerializer:
    def loads(self, data):
        return data

    def dumps(self, data):
        return data


class DummyCacheManager:
    def __init__(self, hit=True):
        self.hit = hit

    def get_or_cache(self, get_result, prefixes, dependencies):
        self.get_result = get_result
        self.prefixes = prefixes
        self.dependencies = dependencies
        if self.hit:
            return CacheResult.hit('key', Response())
        else:
            return CacheResult.miss('key', Response())



import logging

from zope.interface import classImplements

from pyramid_caching.interfaces import IKeyVersioner

log = logging.getLogger(__name__)


def includeme(config):
    config.add_directive('get_key_version_client', get_key_version_client,
                         action_wrap=False)
    config.add_directive('add_key_version_client', add_key_version_client)


def get_key_version_client(config_or_request):
    return config_or_request.registry.getUtility(IKeyVersioner)


def add_key_version_client(config, key_versioner):
    if not IKeyVersioner.implementedBy(key_versioner.__class__):
        log.debug('assuming %r implements %r',
                  key_versioner.__class__, IKeyVersioner)
        classImplements(key_versioner.__class__, IKeyVersioner)

    def register():
        log.debug('registering KeyVersioner %r', key_versioner)
        config.registry.registerUtility(key_versioner, IKeyVersioner)

    config.action((__name__, 'key_version_client'), register, order=0)



from pyramid_caching.tests.functional import Base


class Test(Base):

    def setUp(self):
        # creating app to setup cache and decorate stuff.
        self.app
        from example.model import User, Session
        session = Session()
        session.add(User(id=1, name='Bob'))
        session.commit()

        self.cache_client.client.delete('cache')

    def tearDown(self):
        self.cache_client.client.delete('cache')

    def test_cache_view(self):
        self.app.get('/users/1')
        self.app.get('/users/1')
        self.app.get('/users/1')

        # TODO: test if cache was enabled

    def test_inhibited_cache_view(self):
        self.key_versioner_client.client.set('cache', 'off')

        self.app.get('/users/1')
        self.app.get('/users/1')
        self.app.get('/users/1')

        # TODO: test if cache was disabled



import mock

from zope.interface import implementer

from pyramid_caching.interfaces import ISerializer
from pyramid_caching.tests.functional import Base as TestCase


class MetricsFunctionalTests(TestCase):
    def setUp(self):
        from example.model import User, Session
        registry = self.config.registry
        registry.registerUtility(TestSerializer(), ISerializer)
        self.app
        session = Session()
        session.add(User(id=1, name='Bob'))
        session.commit()

    @mock.patch('pyramid_metrics.utility.StatsClient')
    def test_cache_miss(self, m_statsd):
        self.app.get('/users/1')
        m_incr = m_statsd.return_value.incr
        m_incr.assert_called_once_with(
            'cache.miss.example.views:get_user', count=1)

    @mock.patch('pyramid_metrics.utility.StatsClient')
    def test_cache_hit(self, m_statsd):
        self.app.get('/users/1')
        self.app.get('/users/1')

        m_incr = m_statsd.return_value.incr
        m_incr.assert_has_calls([
            mock.call('cache.miss.example.views:get_user', count=1),
            mock.call('cache.hit.example.views:get_user', count=1),
            ])


@implementer(ISerializer)
class TestSerializer(object):
    def loads(self, data):
        from pyramid.response import Response
        return Response()

    def dumps(self, data):
        return "data"



from zope.interface import implementer

from pyramid_caching.interfaces import ICacheHit, ICacheMiss


@implementer(ICacheHit)
class ViewCacheHit(object):
    """An instance of this class is emitted as an event when the cache manager
    could successfully retrieve a Response object from the cache.
    """
    def __init__(self, cache_key, request):
        self.cache_key = cache_key
        self.request = request


@implementer(ICacheMiss)
class ViewCacheMiss(object):
    """An instance of this class is emitted as an event when the cache manager
    could not retrieve the response from the cache and the object instance was
    created by calling the view method.
    """
    def __init__(self, cache_key, request):
        self.cache_key = cache_key
        self.request = request



import logging

log = logging.getLogger(__name__)


def includeme(config):  # pragma no cover
    config.include('pyramid_caching')
    try:
        config.include('pyramid_caching.ext.metrics')
    except ImportError:
        log.warning('Cache performance metrics library not available.')
    config.include('pyramid_caching.ext.redis')
    config.include('.model')
    config.include('.views')



from setuptools import setup

setup(
    setup_requires=['d2to1'],
    d2to1=True,
    test_suite="nose.collector",
    extras_require={
        'redis': ['redis', 'hiredis'],
        'metrics': ['pyramid_metrics >= 0.1.5'],
    },
)



from pyramid.decorator import reify
from . import Base


class Test(Base):
    def setUp(self):
        # creating app to setup sqla session.
        self.app

    def tearDown(self):
        from example.model import User
        for user in self.sqla_session.query(User).all():
            self.sqla_session.delete(user)
        try:
            self.sqla_session.commit()
        except:  # pragma no cover
            pass

    @reify
    def sqla_session(self):
        from example.model import Session
        session = Session()
        self.addCleanup(session.close)
        self.addCleanup(delattr, self, 'sqla_session')
        return session

    def test_incr(self):
        from example.model import User
        user = User()

        key_cls_0 = self.versioner.get_multi_keys([User])
        key_obj_0 = self.versioner.get_multi_keys([user])

        self.versioner.incr(user)

        key_cls_1 = self.versioner.get_multi_keys([User])
        key_obj_1 = self.versioner.get_multi_keys([user])

        self.assertNotEqual(key_cls_1, key_cls_0)
        self.assertNotEqual(key_obj_1, key_obj_0)

    def test_unicity_simple_pk(self):
        from example.model import User

        user0 = User(name=u'Bob Marley')
        user1 = User(name=u'Peter Tosh')

        self.sqla_session.add(user0)
        self.sqla_session.add(user1)
        self.sqla_session.commit()

        self.assertNotEqual(
            self.versioner.get_multi_keys([user0]),
            self.versioner.get_multi_keys([user1]),
            )

    def test_unicity_composite_pk(self):
        from example.model import User, UserNote

        msg0 = UserNote(id=11, content='I ray')
        msg1 = UserNote(id=22, content='Jah')
        user = User(name=u'Bob Marley')

        user.notes = [msg0, msg1]

        self.sqla_session.add(user)
        self.sqla_session.commit()

        self.assertNotEqual(
            self.versioner.get_multi_keys([msg0]),
            self.versioner.get_multi_keys([msg1]),
            )

    def test_auto_incr(self):
        from example.model import User, UserNote

        # create user and user notes
        user = User(id=1, name=u'Bob')
        msg0 = UserNote(id=11, content='I ray')
        msg1 = UserNote(id=22, content='Jah')
        user.notes = [msg0, msg1]

        self.sqla_session.add(user)

        self.sqla_session.commit()

        key_user_v0 = self.versioner.get_multi_keys([user])
        key_msg0_v0 = self.versioner.get_multi_keys([msg0])
        key_msg1_v0 = self.versioner.get_multi_keys([msg1])

        # modify user and msg0
        user.name = 'Bob Marley'
        msg0.content = 'I RAY'

        self.sqla_session.commit()

        key_user_v1 = self.versioner.get_multi_keys([user])
        key_msg0_v1 = self.versioner.get_multi_keys([msg0])

        # keys are automaticaly changed
        self.assertNotEqual(key_user_v0, key_user_v1)
        self.assertNotEqual(key_msg0_v0, key_msg0_v1)
        # not msg1
        self.assertEqual(key_msg1_v0, self.versioner.get_multi_keys([msg1]))

        # we can get a key via a new object instance
        self.assertEqual(
            key_msg0_v1,
            self.versioner.get_multi_keys([UserNote(id=11, user_id=user.id)]),
            )

        user = self.sqla_session.query(User).get(1)
        # remove msg0
        user.notes.pop(0)
        user_id = user.id
        self.sqla_session.commit()

        # msg0 key has changed
        key_msg0_v2 = self.versioner.get_multi_keys([UserNote(id=11,
                                                    user_id=user_id)])
        self.assertNotEqual(key_msg0_v1, key_msg0_v2)

        # delete user which cascade delete msg1
        user = self.sqla_session.query(User).get(1)
        self.sqla_session.delete(user)
        self.sqla_session.commit()

        self.assertNotEqual(key_user_v1,
                            self.versioner.get_multi_keys([user]))

        self.assertNotEqual(key_msg1_v0, self.versioner.get_multi_keys([msg1]))
